import copy
import tempfile
from unittest.mock import Mock, patch

from octue.cloud.deployment.google.cloud_run.deployer import DEFAULT_CLOUD_RUN_DOCKERFILE_URL, CloudRunDeployer
from octue.exceptions import DeploymentError
from tests.base import BaseTestCase


OCTUE_CONFIGURATION = {
    "services": [
        {
            "name": "test-service",
            "repository_name": "test-repository",
            "repository_owner": "octue",
            "project_name": "test-project",
            "region": "europe-west2",
            "branch_pattern": "my-branch",
        }
    ]
}

SERVICE = OCTUE_CONFIGURATION["services"][0]
GET_SUBSCRIPTIONS_METHOD_PATH = "octue.cloud.deployment.google.cloud_run.deployer.Topic.get_subscriptions"
SERVICE_ID = "octue.services.0df08f9f-30ad-4db3-8029-ea584b4290b7"
EXPECTED_IMAGE_NAME = f"eu.gcr.io/{SERVICE['project_name']}/{SERVICE['repository_name']}/{SERVICE['name']}:$SHORT_SHA"

EXPECTED_CLOUD_BUILD_CONFIGURATION = {
    "steps": [
        {
            "id": "Get default Octue Dockerfile",
            "name": "alpine:latest",
            "args": ["wget", DEFAULT_CLOUD_RUN_DOCKERFILE_URL],
        },
        {
            "id": "Build image",
            "name": "gcr.io/cloud-builders/docker",
            "args": [
                "-c",
                f"docker build '-t' {EXPECTED_IMAGE_NAME!r} . '-f' Dockerfile",
            ],
            "entrypoint": "bash",
        },
        {
            "id": "Push image",
            "name": "gcr.io/cloud-builders/docker",
            "args": ["push", EXPECTED_IMAGE_NAME],
        },
        {
            "id": "Deploy image to Google Cloud Run",
            "name": "gcr.io/google.com/cloudsdktool/cloud-sdk:slim",
            "entrypoint": "gcloud",
            "args": [
                "run",
                "services",
                "update",
                "test-service",
                "--platform=managed",
                f"--image={EXPECTED_IMAGE_NAME}",
                f"--region={SERVICE['region']}",
                "--memory=128Mi",
                "--cpu=1",
                f"--set-env-vars=SERVICE_ID={SERVICE_ID},SERVICE_NAME={SERVICE['name']}",
                "--timeout=3600",
                "--concurrency=10",
                "--min-instances=0",
                "--max-instances=10",
                "--ingress=internal",
            ],
        },
    ],
    "images": [EXPECTED_IMAGE_NAME],
}

EXPECTED_BUILD_TRIGGER_CREATION_COMMAND = [
    "gcloud",
    f"--project={SERVICE['project_name']}",
    "beta",
    "builds",
    "triggers",
    "create",
    "github",
    f"--name={SERVICE['name']}",
    f"--repo-name={SERVICE['repository_name']}",
    f"--repo-owner={SERVICE['repository_owner']}",
    f"--description=Build the {SERVICE['name']!r} service and deploy it to Cloud Run.",
    f"--branch-pattern={SERVICE['branch_pattern']}",
]


class TestCloudRunDeployer(BaseTestCase):
    def test_generate_cloud_build_configuration(self):
        """Test that a correct Google Cloud Build configuration is generated from the given `octue.yaml` file."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = CloudRunDeployer(octue_configuration_path, service_id=SERVICE_ID)
            deployer._generate_cloud_build_configuration()

        self.assertEqual(deployer.generated_cloud_build_configuration, EXPECTED_CLOUD_BUILD_CONFIGURATION)

    def test_generate_cloud_build_configuration_with_custom_dockerfile(self):
        """Test that a correct Google Cloud Build configuration is generated from the given `octue.yaml` file when a
        dockerfile path is given.
        """
        octue_configuration_with_custom_dockerfile = copy.deepcopy(OCTUE_CONFIGURATION)
        octue_configuration_with_custom_dockerfile["services"][0]["dockerfile_path"] = "path/to/Dockerfile"

        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(
                octue_configuration_with_custom_dockerfile,
                temporary_directory,
            )

            deployer = CloudRunDeployer(octue_configuration_path, service_id=SERVICE_ID)
            deployer._generate_cloud_build_configuration()

        # Expect the extra "Get default Octue Dockerfile" step to be absent and the given Dockerfile path to be
        # provided in the first step.
        expected_cloud_build_configuration = copy.deepcopy(EXPECTED_CLOUD_BUILD_CONFIGURATION)
        expected_cloud_build_configuration["steps"] = expected_cloud_build_configuration["steps"][1:]

        expected_cloud_build_configuration["steps"][0]["args"][1] = (
            f"docker build '-t' '{EXPECTED_IMAGE_NAME}' . '-f' "
            f"{octue_configuration_with_custom_dockerfile['services'][0]['dockerfile_path']}"
        )

        self.assertEqual(deployer.generated_cloud_build_configuration, expected_cloud_build_configuration)

    def test_generate_cloud_build_configuration_with_build_secrets(self):
        """Test generating a Cloud Build configuration with build secrets."""
        secret_name = "MY_BIG_SECRET"

        octue_configuration_with_build_secrets = copy.deepcopy(OCTUE_CONFIGURATION)
        service = octue_configuration_with_build_secrets["services"][0]
        service["secrets"] = {"build": [secret_name]}

        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(
                octue_configuration_with_build_secrets,
                temporary_directory,
            )

            deployer = CloudRunDeployer(octue_configuration_path, service_id=SERVICE_ID)
            deployer._generate_cloud_build_configuration()

        expected_cloud_build_configuration = copy.deepcopy(EXPECTED_CLOUD_BUILD_CONFIGURATION)

        expected_cloud_build_configuration["availableSecrets"] = {
            "secretManager": [
                {
                    "versionName": f'projects/{service["project_name"]}/secrets/{secret_name}/versions/latest',
                    "env": secret_name,
                }
            ]
        }

        expected_cloud_build_configuration["steps"][1]["args"][
            1
        ] = f"docker build '-t' {EXPECTED_IMAGE_NAME!r} --build-arg={secret_name}=$${secret_name} . '-f' Dockerfile"

        expected_cloud_build_configuration["steps"][1]["secretEnv"] = [secret_name]

        self.assertEqual(deployer.generated_cloud_build_configuration, expected_cloud_build_configuration)

    def test_deploy(self):
        """Test that the build trigger creation, build and deployment, and Eventarc run trigger creation are requested
        correctly.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = CloudRunDeployer(octue_configuration_path, service_id=SERVICE_ID)

            with patch("subprocess.run", return_value=Mock(returncode=0)) as mock_run:
                with patch("octue.cloud.deployment.google.cloud_run.deployer.Topic.create"):
                    with patch(GET_SUBSCRIPTIONS_METHOD_PATH, return_value=["test-service"]):
                        with patch("octue.cloud.deployment.google.cloud_run.deployer.Subscription"):
                            with patch("octue.cloud.deployment.google.cloud_run.deployer.Service"):
                                mock_build_id = "my-build-id"
                                with patch(
                                    "json.loads",
                                    return_value={
                                        "metadata": {"build": {"images": ["my-image"], "id": mock_build_id}},
                                        "status": "SUCCESS",
                                    },
                                ):
                                    temporary_file = tempfile.NamedTemporaryFile(delete=False)

                                    with patch("tempfile.NamedTemporaryFile", return_value=temporary_file):
                                        deployer.deploy()

            # Test the build trigger creation request.
            self.assertEqual(
                mock_run.call_args_list[0].args[0],
                EXPECTED_BUILD_TRIGGER_CREATION_COMMAND + [f"--inline-config={temporary_file.name}"],
            )

            # Test the build trigger run request.
            self.assertEqual(
                mock_run.call_args_list[1].args[0],
                [
                    "gcloud",
                    f"--project={SERVICE['project_name']}",
                    "--format=json",
                    "beta",
                    "builds",
                    "triggers",
                    "run",
                    SERVICE["name"],
                    "--branch=my-branch",
                ],
            )

            # Test waiting for the build trigger run to complete.
            self.assertEqual(
                mock_run.call_args_list[2].args[0],
                [
                    "gcloud",
                    f'--project={SERVICE["project_name"]}',
                    "--format=json",
                    "builds",
                    "describe",
                    mock_build_id,
                ],
            )

            # Test setting the Cloud Run service to accept unauthenticated requests.
            self.assertEqual(
                mock_run.call_args_list[3].args[0],
                [
                    "gcloud",
                    f'--project={SERVICE["project_name"]}',
                    "run",
                    "services",
                    "add-iam-policy-binding",
                    SERVICE["name"],
                    f'--region={SERVICE["region"]}',
                    "--member=allUsers",
                    "--role=roles/run.invoker",
                ],
            )

            # Test the Eventarc run trigger creation request.
            self.assertEqual(
                mock_run.call_args_list[4].args[0],
                [
                    "gcloud",
                    f'--project={SERVICE["project_name"]}',
                    "beta",
                    "eventarc",
                    "triggers",
                    "create",
                    f'{SERVICE["name"]}-trigger',
                    "--matching-criteria=type=google.cloud.pubsub.topic.v1.messagePublished",
                    f"--destination-run-service={SERVICE['name']}",
                    f"--location={SERVICE['region']}",
                    f"--transport-topic={SERVICE_ID}",
                ],
            )

    def test_create_build_trigger_with_update(self):
        """Test that creating a build trigger for a service when one already exists and the deployer is in `update`
        mode results in the existing trigger being deleted and recreated.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = CloudRunDeployer(octue_configuration_path)
            deployer._generate_cloud_build_configuration()

            temporary_file = tempfile.NamedTemporaryFile(delete=False)

            with patch("tempfile.NamedTemporaryFile", return_value=temporary_file):
                with patch(
                    "octue.cloud.deployment.google.cloud_run.deployer.CloudRunDeployer._run_command",
                    side_effect=[DeploymentError("already exists"), None, None],
                ) as mock_run_command:
                    with patch("builtins.print") as mock_print:
                        deployer._create_build_trigger(update=True)

        self.assertEqual(mock_print.call_args[0][0], "recreated.")

        expected_build_trigger_creation_command = EXPECTED_BUILD_TRIGGER_CREATION_COMMAND + [
            f"--inline-config={temporary_file.name}"
        ]

        # Test the build trigger creation request.
        self.assertEqual(mock_run_command.call_args_list[0].args[0], expected_build_trigger_creation_command)

        # Test that trigger deletion is requested.
        self.assertEqual(
            mock_run_command.call_args_list[1].args[0],
            [
                "gcloud",
                f"--project={SERVICE['project_name']}",
                "beta",
                "builds",
                "triggers",
                "delete",
                SERVICE["name"],
            ],
        )

        # Test the build trigger creation request is retried.
        self.assertEqual(mock_run_command.call_args_list[2].args[0], expected_build_trigger_creation_command)

    def test_create_eventarc_run_trigger_with_update(self):
        """Test that creating an Eventarc run trigger for a service when one already exists and the deployer is in
        `update` mode results in the Eventarc subscription update being skipped and an "already exists" message being
        printed.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = CloudRunDeployer(octue_configuration_path)

            with patch("octue.cloud.deployment.google.cloud_run.deployer.Topic.create"):
                with patch(
                    "octue.cloud.deployment.google.cloud_run.deployer.CloudRunDeployer._run_command",
                    side_effect=DeploymentError("already exists"),
                ):
                    with patch(GET_SUBSCRIPTIONS_METHOD_PATH) as mock_get_subscriptions:
                        with patch("builtins.print") as mock_print:
                            deployer._create_eventarc_run_trigger(update=True)

        mock_get_subscriptions.assert_not_called()
        self.assertEqual(mock_print.call_args[0][0], "already exists.")



import concurrent.futures
import json
import os
import warnings

from octue import definitions
from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.exceptions import InvalidInputException
from octue.mixins import Hashable, Identifiable, Labelable, Pathable, Serialisable, Taggable
from octue.resources.datafile import Datafile
from octue.resources.filter_containers import FilterSet
from octue.resources.label import LabelSet
from octue.resources.tag import TagDict


DATAFILES_DIRECTORY = "datafiles"


class Dataset(Labelable, Taggable, Serialisable, Pathable, Identifiable, Hashable):
    """A representation of a dataset, containing files, labels, etc

    This is used to read a list of files (and their associated properties) into octue analysis, or to compile a
    list of output files (results) and their properties that will be sent back to the octue system.

    :param iter(dict|octue.resources.datafile.Datafile) files: the files belonging to the dataset
    :return None:
    """

    _ATTRIBUTES_TO_HASH = ("files",)
    _SERIALISE_FIELDS = "files", "name", "labels", "tags", "id", "path"

    def __init__(self, files=None, name=None, id=None, path=None, path_from=None, tags=None, labels=None, **kwargs):
        super().__init__(name=name, id=id, tags=tags, labels=labels, path=path, path_from=path_from)

        # TODO The decoders aren't being used; utils.decoders.OctueJSONDecoder should be used in twined
        #  so that resources get automatically instantiated.
        #  Add a proper `decoder` argument  to the load_json utility in twined so that datasets, datafiles and manifests
        #  get initialised properly, then remove this hackjob.
        self.files = FilterSet()
        self._cloud_path = None

        for file in files or []:
            if isinstance(file, Datafile):
                self.files.add(file)
            else:
                self.files.add(Datafile.deserialise(file, path_from=self))

        self.__dict__.update(**kwargs)

    def __iter__(self):
        yield from self.files

    def __len__(self):
        return len(self.files)

    @classmethod
    def from_local_directory(cls, path_to_directory, recursive=False, **kwargs):
        """Instantiate a Dataset from the files in the given local directory.

        :param str path_to_directory: path to a local directory
        :param bool recursive: if `True`, include all files in the directory's subdirectories recursively
        :param kwargs: other keyword arguments for the `Dataset` instantiation
        :return Dataset:
        """
        datafiles = FilterSet()

        for level, (directory_path, _, filenames) in enumerate(os.walk(path_to_directory)):
            for filename in filenames:

                if not recursive and level > 0:
                    break

                datafiles.add(Datafile(path=os.path.join(directory_path, filename)))

        return Dataset(path=path_to_directory, files=datafiles, **kwargs)

    @classmethod
    def from_cloud(
        cls,
        project_name=None,
        cloud_path=None,
        bucket_name=None,
        path_to_dataset_directory=None,
        recursive=False,
    ):
        """Instantiate a Dataset from Google Cloud storage. Either (`bucket_name` and `path_to_dataset_directory`) or
        `cloud_path` must be provided.

        :param str|None cloud_path: full path to dataset directory in cloud storage (e.g. `gs://bucket_name/path/to/dataset`)
        :param str|None bucket_name: name of bucket dataset is stored in
        :param str|None path_to_dataset_directory: path to dataset directory (containing dataset's files) in cloud (e.g. `path/to/dataset`)
        :param bool recursive: if `True`, include in the dataset all files in the subdirectories recursively contained in the dataset directory
        :return Dataset:
        """
        if project_name:
            warnings.warn(
                message="The `project_name` parameter is no longer needed and will be removed soon.",
                category=DeprecationWarning,
            )

        if cloud_path:
            bucket_name = storage.path.split_bucket_name_from_gs_path(cloud_path)[0]
        else:
            cloud_path = storage.path.generate_gs_path(bucket_name, path_to_dataset_directory)

        dataset_metadata = cls._get_dataset_metadata(cloud_path=cloud_path)

        if dataset_metadata:
            return Dataset(
                id=dataset_metadata.get("id"),
                name=dataset_metadata.get("name"),
                path=cloud_path,
                tags=TagDict(dataset_metadata.get("tags", {})),
                labels=LabelSet(dataset_metadata.get("labels", [])),
                files=[Datafile(path=path) for path in dataset_metadata["files"]],
            )

        datafiles = FilterSet(
            Datafile(path=storage.path.generate_gs_path(bucket_name, blob.name))
            for blob in GoogleCloudStorageClient().scandir(cloud_path, recursive=recursive)
        )

        dataset = Dataset(path=cloud_path, files=datafiles)
        dataset._upload_dataset_metadata(cloud_path)
        return dataset

    def to_cloud(self, project_name=None, cloud_path=None, bucket_name=None, output_directory=None):
        """Upload a dataset to a cloud location. Either (`bucket_name` and `output_directory`) or `cloud_path` must be
        provided.

        :param str|None cloud_path: full cloud storage path to store dataset at (e.g. `gs://bucket_name/path/to/dataset`)
        :param str|None bucket_name: name of bucket to store dataset in
        :param str|None output_directory: path to output directory in cloud storage (e.g. `path/to/dataset`)
        :return str: cloud path for dataset
        """
        if project_name:
            warnings.warn(
                message="The `project_name` parameter is no longer needed and will be removed soon.",
                category=DeprecationWarning,
            )

        if not cloud_path:
            cloud_path = storage.path.generate_gs_path(bucket_name, output_directory, self.name)

        files_and_paths = []

        for datafile in self.files:
            datafile_path_relative_to_dataset = datafile.path.split(self.path)[-1].strip(os.path.sep).strip("/")
            files_and_paths.append(
                (
                    datafile,
                    storage.path.join(cloud_path, *datafile_path_relative_to_dataset.split(os.path.sep)),
                )
            )

        def upload(iterable_element):
            """Upload a datafile to the given cloud path.

            :param tuple(octue.resources.datafile.Datafile, str) iterable_element:
            :return None:
            """
            datafile = iterable_element[0]
            cloud_path = iterable_element[1]
            datafile.to_cloud(cloud_path=cloud_path)

        # Use multiple threads to significantly speed up file uploads by reducing latency.
        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(upload, files_and_paths)

        self._upload_dataset_metadata(cloud_path)
        self._cloud_path = cloud_path
        return cloud_path

    @property
    def name(self):
        """Get the name of the dataset

        :return str:
        """
        return self._name or os.path.split(os.path.abspath(os.path.split(self.path)[-1]))[-1]

    @property
    def cloud_path(self):
        """Get the cloud path of the dataset.

        :return str|None:
        """
        return self._cloud_path

    @cloud_path.setter
    def cloud_path(self, path):
        """Set the cloud path of the dataset.

        :param str|None path:
        :return None:
        """
        if path is None:
            self._cloud_path = None
            return

        self.to_cloud(cloud_path=path)

    @property
    def all_files_are_in_cloud(self):
        """Do all the files of the dataset exist in the cloud?

        :return bool:
        """
        return all(file.exists_in_cloud for file in self.files)

    def add(self, *args, **kwargs):
        """Add a data/results file to the manifest.

        Usage:
            my_file = octue.DataFile(...)
            my_manifest.add(my_file)

            # or more simply
            my_manifest.add(**{...}) which implicitly creates the datafile from the starred list of input arguments
        """
        if len(args) > 1:
            # Recurse to allow addition of many files at once
            for arg in args:
                self.add(arg, **kwargs)
        elif len(args) > 0:
            if not isinstance(args[0], Datafile):
                raise InvalidInputException(
                    'Object "{}" must be of class Datafile to add it to a Dataset'.format(args[0])
                )
            self.files.add(args[0])

        else:
            # Add a single file, constructed by passing the arguments through to DataFile()
            self.files.add(Datafile(**kwargs))

    def get_file_by_label(self, label):
        """Get a single datafile from a dataset by filtering for files with the provided label.

        :param str label: the label to filter for
        :raise octue.exceptions.UnexpectedNumberOfResultsException: if zero or more than one results satisfy the filters
        :return octue.resources.datafile.DataFile:
        """
        return self.files.one(labels__contains=label)

    def download_all_files(self, local_directory=None):
        """Download all files in the dataset into the given local directory. If no path to a local directory is given,
        the files will be downloaded to temporary locations.

        :param str|None local_directory:
        :return None:
        """
        files_and_paths = []

        for datafile in self.files:
            if local_directory:
                path_relative_to_dataset = storage.path.relpath(datafile.cloud_path, self.path)
                local_path = os.path.abspath(os.path.join(local_directory, *path_relative_to_dataset.split("/")))
            else:
                local_path = None

            files_and_paths.append((datafile, local_path))

        def download(iterable_element):
            """Download a datafile to the given path.

            :param tuple(octue.resources.datafile.Datafile, str) iterable_element:
            :return str:
            """
            datafile = iterable_element[0]
            local_path = iterable_element[1]
            return datafile.download(local_path=local_path)

        # Use multiple threads to significantly speed up files downloads by reducing latency.
        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(download, files_and_paths)

    @staticmethod
    def _get_dataset_metadata(cloud_path):
        """Get the metadata for the given dataset if a dataset metadata file has previously been uploaded.

        :param str cloud_path: the path to the dataset cloud directory
        :return dict: the dataset metadata
        """
        storage_client = GoogleCloudStorageClient()
        metadata_file_path = storage.path.join(cloud_path, definitions.DATASET_METADATA_FILENAME)

        if not storage_client.exists(cloud_path=metadata_file_path):
            return {}

        return json.loads(storage_client.download_as_string(cloud_path=metadata_file_path))

    def _upload_dataset_metadata(self, cloud_path):
        """Upload a metadata file representing the dataset to the given cloud location.

        :param str cloud_path: the path to the dataset cloud directory
        :return None:
        """
        serialised_dataset = self.to_primitive()
        serialised_dataset["files"] = sorted(datafile.cloud_path for datafile in self.files)
        del serialised_dataset["path"]

        GoogleCloudStorageClient().upload_from_string(
            string=json.dumps(serialised_dataset),
            cloud_path=storage.path.join(cloud_path, definitions.DATASET_METADATA_FILENAME),
        )



from octue import exceptions
from octue.mixins import Taggable
from octue.resources.tag import TagDict

from ..base import BaseTestCase


class MyTaggable(Taggable):
    pass


class TaggableTestCase(BaseTestCase):
    def test_instantiates(self):
        """Ensures the class instantiates without arguments."""
        taggable = Taggable()
        self.assertEqual(taggable.tags, {})

    def test_instantiating_with_no_tags(self):
        """Test that instantiating a Taggable with no tags results in an empty TagDict on the tags attribute."""
        self.assertEqual(MyTaggable().tags, TagDict())

    def test_fails_to_instantiates_with_non_iterable(self):
        """Test that instantiation with a non-iterable fails."""

        class NoIter:
            pass

        with self.assertRaises(TypeError):
            MyTaggable(tags=NoIter())

    def test_instantiates_with_dict(self):
        """Test instantiation with a dictionary works."""
        tags = {"height": 9, "width": 8.7, "depth": 100}
        taggable = MyTaggable(tags=tags)
        self.assertEqual(taggable.tags, tags)

    def test_instantiates_with_tag_dict(self):
        """Test instantiation with a TagDict."""
        taggable_1 = MyTaggable(tags={"a": 2})
        self.assertIsInstance(taggable_1.tags, TagDict)
        taggable_2 = MyTaggable(tags=taggable_1.tags)
        self.assertFalse(taggable_1 is taggable_2)

    def test_setting_tags_overwrites_previous_tags(self):
        """Ensure tags can be overwritten with new ones."""
        taggable = MyTaggable(tags={"a": 1, "b": 2})
        taggable.tags = {"c": 3, "d": 4}
        self.assertEqual(taggable.tags, {"c": 3, "d": 4})

    def test_add_valid_tags(self):
        """Ensures adding valid tags works."""
        taggable = MyTaggable()

        taggable.add_tags({"a_valid_tag": "blah"})
        taggable.add_tags({"a1829tag": "blah"})
        taggable.add_tags({"1829": "blah", "number_1829": "blah"})  # Add multiple tags at once.

        self.assertEqual(
            taggable.tags,
            {"a_valid_tag": "blah", "a1829tag": "blah", "1829": "blah", "number_1829": "blah"},
        )

    def test_add_tags_via_kwargs(self):
        """Test tags can be added via kwargs."""
        taggable = MyTaggable()
        taggable.add_tags(hello="blib", hi="glib")
        self.assertEqual(taggable.tags, {"hello": "blib", "hi": "glib"})

    def test_adding_mixture_of_valid_and_invalid_tags_fails_completely(self):
        """Ensure that adding a variety of tags, some of which are invalid, doesn't partially add the set including the
        invalid tags to the object.
        """
        taggable = MyTaggable()
        taggable.add_tags({"first_valid_should_be_added": "hello"})

        with self.assertRaises(exceptions.InvalidTagException):
            taggable.add_tags({"second_valid_should_not_be_added_because": 1, "_the_third_is_invalid:": 2})

        self.assertEqual(taggable.tags, {"first_valid_should_be_added": "hello"})



import base64
import json
import uuid
from unittest import TestCase, mock

from octue.cloud.deployment.google.cloud_run import flask_app


flask_app.app.testing = True


class TestFlaskApp(TestCase):
    def test_post_to_index_with_no_payload_results_in_400_error(self):
        """Test that a 400 (bad request) error code is returned if no payload is sent to the Flask endpoint."""
        with flask_app.app.test_client() as client:
            response = client.post("/", json={"deliveryAttempt": 1})
            self.assertEqual(response.status_code, 400)

    def test_post_to_index_with_invalid_payload_results_in_400_error(self):
        """Test that a 400 (bad request) error code is returned if an invalid payload is sent to the Flask endpoint."""
        with flask_app.app.test_client() as client:
            response = client.post("/", json={"some": "data", "deliveryAttempt": 1})
            self.assertEqual(response.status_code, 400)

            response = client.post("/", json={"message": "data", "deliveryAttempt": 1})
            self.assertEqual(response.status_code, 400)

    def test_post_to_index_with_valid_payload(self):
        """Test that the Flask endpoint returns a 204 (ok, no content) response to a valid payload."""
        with flask_app.app.test_client() as client:
            with mock.patch("octue.cloud.deployment.google.cloud_run.flask_app.answer_question"):

                response = client.post(
                    "/",
                    json={
                        "deliveryAttempt": 1,
                        "subscription": "projects/my-project/subscriptions/my-subscription",
                        "message": {
                            "data": base64.b64encode(
                                json.dumps({"input_values": [1, 2, 3], "input_manifest": None}).encode()
                            ).decode(),
                            "attributes": {"question_uuid": str(uuid.uuid4()), "forward_logs": "1"},
                        },
                    },
                )

                self.assertEqual(response.status_code, 204)

    def test_redelivered_questions_are_acknowledged_and_ignored(self):
        """Test that redelivered questions are acknowledged and then ignored."""
        with flask_app.app.test_client() as client:
            with mock.patch(
                "octue.cloud.deployment.google.cloud_run.flask_app.answer_question"
            ) as mock_answer_question:

                response = client.post(
                    "/",
                    json={
                        "deliveryAttempt": 2,
                        "subscription": "projects/my-project/subscriptions/my-subscription",
                        "message": {
                            "data": {},
                            "attributes": {"question_uuid": str(uuid.uuid4()), "forward_logs": "1"},
                        },
                    },
                )

        self.assertEqual(response.status_code, 204)
        mock_answer_question.assert_not_called()



import copy
import tempfile
from unittest.mock import Mock, patch

from apache_beam.runners.dataflow.internal.apiclient import DataflowJobAlreadyExistsError

from octue.cloud.deployment.google.dataflow.deployer import (
    DEFAULT_DATAFLOW_DOCKERFILE_URL,
    DEFAULT_DATAFLOW_TEMPORARY_FILES_LOCATION,
    DEFAULT_SETUP_FILE_PATH,
    OCTUE_SDK_PYTHON_IMAGE_URI,
    DataflowDeployer,
)
from octue.exceptions import DeploymentError
from tests.base import BaseTestCase


OCTUE_CONFIGURATION = {
    "services": [
        {
            "name": "test-service",
            "repository_name": "test-repository",
            "repository_owner": "octue",
            "project_name": "test-project",
            "region": "europe-west2",
            "branch_pattern": "my-branch",
            "service_account_email": "account@domain.com",
            "maximum_instances": 30,
            "machine_type": "e2-standard-2",
        }
    ]
}

SERVICE = OCTUE_CONFIGURATION["services"][0]

OCTUE_CONFIGURATION_WITH_CLOUD_BUILD_PATH = {
    "services": [{**copy.copy(SERVICE), "cloud_build_configuration_path": "cloudbuild.yaml"}]
}

SERVICE_ID = "octue.services.0df08f9f-30ad-4db3-8029-ea584b4290b7"

EXPECTED_IMAGE_NAME = (
    f"eu.gcr.io/{SERVICE['project_name']}/{SERVICE['repository_name']}/" f"{SERVICE['name']}:$SHORT_SHA"
)

EXPECTED_CLOUD_BUILD_CONFIGURATION = {
    "steps": [
        {
            "id": "Get default Octue Dockerfile",
            "name": "alpine:latest",
            "args": ["wget", DEFAULT_DATAFLOW_DOCKERFILE_URL],
        },
        {
            "id": "Build image",
            "name": "gcr.io/cloud-builders/docker",
            "entrypoint": "bash",
            "args": [
                "-c",
                (
                    f"docker build '-t' {EXPECTED_IMAGE_NAME!r} --build-arg=SERVICE_ID={SERVICE_ID} "
                    f"--build-arg=SERVICE_NAME={SERVICE['name']} . '-f' Dockerfile"
                ),
            ],
        },
        {
            "id": "Push image",
            "name": "gcr.io/cloud-builders/docker",
            "args": ["push", EXPECTED_IMAGE_NAME],
        },
        {
            "id": "Deploy Dataflow job",
            "name": OCTUE_SDK_PYTHON_IMAGE_URI,
            "args": [
                "octue-app",
                "deploy",
                "dataflow",
                f"--service-id={SERVICE_ID}",
                "--update",
                "--dataflow-job-only",
                f"--image-uri={EXPECTED_IMAGE_NAME}",
            ],
        },
    ],
    "images": [EXPECTED_IMAGE_NAME],
}

EXPECTED_BUILD_TRIGGER_CREATION_COMMAND = [
    "gcloud",
    f"--project={SERVICE['project_name']}",
    "beta",
    "builds",
    "triggers",
    "create",
    "github",
    f"--name={SERVICE['name']}",
    f"--repo-name={SERVICE['repository_name']}",
    f"--repo-owner={SERVICE['repository_owner']}",
    f"--description=Build the {SERVICE['name']!r} service and deploy it to Dataflow.",
    f"--branch-pattern={SERVICE['branch_pattern']}",
]


class TestDataflowDeployer(BaseTestCase):
    def test_generate_cloud_build_configuration(self):
        """Test that a correct Google Cloud Build configuration is generated from the given `octue.yaml` file."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID)

        deployer._generate_cloud_build_configuration()
        self.assertEqual(deployer.generated_cloud_build_configuration, EXPECTED_CLOUD_BUILD_CONFIGURATION)

    def test_deploy(self):
        """Test that the build trigger creation and run are requested correctly."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID)

            with patch("subprocess.run", return_value=Mock(returncode=0)) as mock_run:
                mock_build_id = "my-build-id"

                with patch(
                    "json.loads",
                    return_value={
                        "metadata": {"build": {"images": [deployer.image_uri_template], "id": mock_build_id}},
                        "status": "SUCCESS",
                    },
                ):
                    temporary_file = tempfile.NamedTemporaryFile(delete=False)

                    with patch("tempfile.NamedTemporaryFile", return_value=temporary_file):
                        deployer.deploy()

            # Test the build trigger creation request.
            self.assertEqual(
                mock_run.call_args_list[0].args[0],
                EXPECTED_BUILD_TRIGGER_CREATION_COMMAND + [f"--inline-config={temporary_file.name}"],
            )

            # Test the build trigger run request.
            self.assertEqual(
                mock_run.call_args_list[1].args[0],
                [
                    "gcloud",
                    f"--project={SERVICE['project_name']}",
                    "--format=json",
                    "beta",
                    "builds",
                    "triggers",
                    "run",
                    SERVICE["name"],
                    "--branch=my-branch",
                ],
            )

            # Test waiting for the build trigger run to complete.
            self.assertEqual(
                mock_run.call_args_list[2].args[0],
                [
                    "gcloud",
                    f'--project={SERVICE["project_name"]}',
                    "--format=json",
                    "builds",
                    "describe",
                    mock_build_id,
                ],
            )

    def test_deploy_with_cloud_build_file_provided(self):
        """Test deploying to Dataflow with a `cloudbuild.yaml` path provided in the `octue.yaml` file"""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(
                OCTUE_CONFIGURATION_WITH_CLOUD_BUILD_PATH,
                temporary_directory,
            )

            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID, image_uri_template="blah")

            with patch("subprocess.run", return_value=Mock(returncode=0)) as mock_run:
                mock_build_id = "my-build-id"

                with patch(
                    "json.loads",
                    return_value={
                        "metadata": {"build": {"images": [deployer.image_uri_template], "id": mock_build_id}},
                        "status": "SUCCESS",
                    },
                ):
                    with patch("octue.cloud.deployment.google.dataflow.pipeline.DataflowRunner"):
                        deployer.deploy()

            # Test the build trigger creation request.
            self.assertEqual(
                mock_run.call_args_list[0].args[0],
                EXPECTED_BUILD_TRIGGER_CREATION_COMMAND
                + [
                    f"--build-config={OCTUE_CONFIGURATION_WITH_CLOUD_BUILD_PATH['services'][0]['cloud_build_configuration_path']}"
                ],
            )

            # Test the build trigger run request.
            self.assertEqual(
                mock_run.call_args_list[1].args[0],
                [
                    "gcloud",
                    f"--project={OCTUE_CONFIGURATION_WITH_CLOUD_BUILD_PATH['services'][0]['project_name']}",
                    "--format=json",
                    "beta",
                    "builds",
                    "triggers",
                    "run",
                    OCTUE_CONFIGURATION_WITH_CLOUD_BUILD_PATH["services"][0]["name"],
                    "--branch=my-branch",
                ],
            )

            # Test waiting for the build trigger run to complete.
            self.assertEqual(
                mock_run.call_args_list[2].args[0],
                [
                    "gcloud",
                    f'--project={SERVICE["project_name"]}',
                    "--format=json",
                    "builds",
                    "describe",
                    mock_build_id,
                ],
            )

    def test_create_streaming_dataflow_job(self):
        """Test creating a streaming dataflow job directly."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID)

            with patch(
                "octue.cloud.deployment.google.dataflow.pipeline.Topic",
                return_value=Mock(path="projects/my-project/topics/my-topic"),
            ):
                with patch("octue.cloud.deployment.google.dataflow.pipeline.DataflowRunner.run_pipeline") as mock_run:
                    deployer.create_streaming_dataflow_job(image_uri="my-image-uri")

            options = mock_run.call_args.kwargs["options"].get_all_options()
            self.assertFalse(options["update"])
            self.assertTrue(options["streaming"])
            self.assertEqual(options["project"], SERVICE["project_name"])
            self.assertEqual(options["job_name"], SERVICE["name"])
            self.assertEqual(options["temp_location"], DEFAULT_DATAFLOW_TEMPORARY_FILES_LOCATION)
            self.assertEqual(options["region"], SERVICE["region"])
            self.assertEqual(options["sdk_container_image"], "my-image-uri")
            self.assertEqual(options["setup_file"], DEFAULT_SETUP_FILE_PATH)

            # Check that "enable_prime" isn't in the Dataflow service options (it should be disabled if a machine type
            # is specified in the octue configuration file).
            self.assertIsNone(options["dataflow_service_options"])

    def test_updating_streaming_dataflow_job(self):
        """Test updating an existing streaming dataflow job."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID)

            with patch(
                "octue.cloud.deployment.google.dataflow.pipeline.Topic",
                return_value=Mock(path="projects/my-project/topics/my-topic"),
            ):
                with patch("octue.cloud.deployment.google.dataflow.pipeline.DataflowRunner.run_pipeline") as mock_run:
                    deployer.create_streaming_dataflow_job(image_uri="my-image-uri", update=True)

            options = mock_run.call_args.kwargs["options"].get_all_options()
            self.assertTrue(options["update"])
            self.assertTrue(options["streaming"])
            self.assertIsNone(options["dataflow_service_options"])
            self.assertEqual(options["project"], SERVICE["project_name"])
            self.assertEqual(options["job_name"], SERVICE["name"])
            self.assertEqual(options["temp_location"], DEFAULT_DATAFLOW_TEMPORARY_FILES_LOCATION)
            self.assertEqual(options["region"], SERVICE["region"])
            self.assertEqual(options["sdk_container_image"], "my-image-uri")
            self.assertEqual(options["setup_file"], DEFAULT_SETUP_FILE_PATH)

    def test_create_streaming_dataflow_job_when_job_does_not_already_exist(self):
        """Test that attempting to deploy an update to a Dataflow job when a job with the name of service does not
        already exist results in the job deployment being retried with `update` set to `False`.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID)

            with patch(
                "octue.cloud.deployment.google.dataflow.pipeline.Topic",
                return_value=Mock(path="projects/my-project/topics/my-topic"),
            ):
                with patch(
                    "octue.cloud.deployment.google.dataflow.pipeline.DataflowRunner.run_pipeline",
                    side_effect=[ValueError, None],
                ) as mock_runner:
                    deployer.create_streaming_dataflow_job(image_uri="my-image-uri", update=True)

            # Check that the first attempt was to update the service.
            first_attempt_options = mock_runner.mock_calls[0].kwargs["options"].get_all_options()
            self.assertTrue(first_attempt_options["update"])

            # Check that the second attempt was to create the service.
            second_attempt_options = mock_runner.mock_calls[1].kwargs["options"].get_all_options()
            self.assertFalse(second_attempt_options["update"])

    def test_deployment_error_raised_if_dataflow_job_already_exists(self):
        """Test that a deployment error is raised if a Dataflow job already exists with the same name as the service."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_path = self._create_octue_configuration_file(OCTUE_CONFIGURATION, temporary_directory)
            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID)

            with patch(
                "octue.cloud.deployment.google.dataflow.pipeline.Topic",
                return_value=Mock(path="projects/my-project/topics/my-topic"),
            ):
                with patch(
                    "octue.cloud.deployment.google.dataflow.pipeline.DataflowRunner.run_pipeline",
                    side_effect=DataflowJobAlreadyExistsError(),
                ):
                    with self.assertRaises(DeploymentError):
                        deployer.create_streaming_dataflow_job(image_uri="my-image-uri", update=True)

    def test_dataflow_prime_enabled_if_machine_type_not_specified(self):
        """Test that Dataflow Prime is enabled if the machine type is not specified in the octue configuration."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            octue_configuration_with_no_machine_type = copy.deepcopy(OCTUE_CONFIGURATION)
            del octue_configuration_with_no_machine_type["services"][0]["machine_type"]

            octue_configuration_path = self._create_octue_configuration_file(
                octue_configuration_with_no_machine_type,
                temporary_directory,
            )

            deployer = DataflowDeployer(octue_configuration_path, service_id=SERVICE_ID)

            with patch(
                "octue.cloud.deployment.google.dataflow.pipeline.Topic",
                return_value=Mock(path="projects/my-project/topics/my-topic"),
            ):
                with patch("octue.cloud.deployment.google.dataflow.pipeline.DataflowRunner.run_pipeline") as mock_run:
                    deployer.create_streaming_dataflow_job(image_uri="my-image-uri")

        self.assertEqual(
            mock_run.call_args.kwargs["options"].get_all_options()["dataflow_service_options"],
            ["enable_prime"],
        )



from unittest.mock import patch

from octue import exceptions
from octue.cloud.pub_sub.message_handler import OrderedMessageHandler
from octue.resources.service_backends import GCPPubSubBackend
from tests import TEST_PROJECT_NAME
from tests.base import BaseTestCase
from tests.cloud.pub_sub.mocks import (
    MockMessagePuller,
    MockPullResponse,
    MockService,
    MockSubscriber,
    MockSubscription,
    MockTopic,
)


BACKEND = GCPPubSubBackend(
    project_name=TEST_PROJECT_NAME, credentials_environment_variable="GOOGLE_APPLICATION_CREDENTIALS"
)


class TestOrderedMessageHandler(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        """Set up the test class with a mock subscription.

        :return None:
        """
        service = MockService(backend=BACKEND)
        mock_topic = MockTopic(name="world", namespace="hello", service=service)
        cls.mock_subscription = MockSubscription(
            name="world",
            topic=mock_topic,
            namespace="hello",
            project_name=TEST_PROJECT_NAME,
            subscriber=MockSubscriber(),
        )

    def _make_order_recording_message_handler(self, message_handling_order):
        """Make a message handler that records the order in which messages were handled to the given list.

        :param list message_handling_order:
        :return callable:
        """

        def message_handler(message):
            message_handling_order.append(message["message_number"])

        return message_handler

    def test_timeout(self):
        """Test that a TimeoutError is raised if message handling takes longer than the given timeout."""
        message_handler = OrderedMessageHandler(
            subscriber=MockSubscriber(),
            subscription=self.mock_subscription,
            message_handlers={
                "test": self._make_order_recording_message_handler([]),
                "finish-test": lambda message: message,
            },
        )

        with patch(
            "octue.cloud.pub_sub.service.OrderedMessageHandler._pull_message",
            new=MockMessagePuller(messages=[{"type": "test", "message_number": 0}]).pull,
        ):
            with self.assertRaises(TimeoutError):
                message_handler.handle_messages(timeout=0)

    def test_unknown_message_type_raises_warning(self):
        """Test that unknown message types result in a warning being logged."""
        message_handler = OrderedMessageHandler(
            subscriber=MockSubscriber(),
            subscription=self.mock_subscription,
            message_handlers={"finish-test": lambda message: message},
        )

        with patch("logging.StreamHandler.emit") as mock_emit:
            message_handler._handle_message({"type": "blah", "message_number": 0})

        self.assertIn("Received a message of unknown type", mock_emit.call_args_list[0][0][0].msg)

    def test_in_order_messages_are_handled_in_order(self):
        """Test that messages received in order are handled in order."""
        message_handling_order = []

        message_handler = OrderedMessageHandler(
            subscriber=MockSubscriber(),
            subscription=self.mock_subscription,
            message_handlers={
                "test": self._make_order_recording_message_handler(message_handling_order),
                "finish-test": lambda message: "This is the result.",
            },
        )

        with patch(
            "octue.cloud.pub_sub.service.OrderedMessageHandler._pull_message",
            new=MockMessagePuller(
                messages=[
                    {"type": "test", "message_number": 0},
                    {"type": "test", "message_number": 1},
                    {"type": "test", "message_number": 2},
                    {"type": "finish-test", "message_number": 3},
                ]
            ).pull,
        ):
            result = message_handler.handle_messages()

        self.assertEqual(result, "This is the result.")
        self.assertEqual(message_handling_order, [0, 1, 2])

    def test_out_of_order_messages_are_handled_in_order(self):
        """Test that messages received out of order are handled in order."""
        message_handling_order = []

        message_handler = OrderedMessageHandler(
            subscriber=MockSubscriber(),
            subscription=self.mock_subscription,
            message_handlers={
                "test": self._make_order_recording_message_handler(message_handling_order),
                "finish-test": lambda message: "This is the result.",
            },
        )

        with patch(
            "octue.cloud.pub_sub.service.OrderedMessageHandler._pull_message",
            new=MockMessagePuller(
                messages=[
                    {"type": "test", "message_number": 1},
                    {"type": "test", "message_number": 2},
                    {"type": "test", "message_number": 0},
                    {"type": "finish-test", "message_number": 3},
                ]
            ).pull,
        ):
            result = message_handler.handle_messages()

        self.assertEqual(result, "This is the result.")
        self.assertEqual(message_handling_order, [0, 1, 2])

    def test_out_of_order_messages_with_end_message_first_are_handled_in_order(self):
        """Test that messages received out of order and with the final message (the message that triggers a value to be
        returned) are handled in order.
        """
        message_handling_order = []

        message_handler = OrderedMessageHandler(
            subscriber=MockSubscriber(),
            subscription=self.mock_subscription,
            message_handlers={
                "test": self._make_order_recording_message_handler(message_handling_order),
                "finish-test": lambda message: "This is the result.",
            },
        )

        with patch(
            "octue.cloud.pub_sub.service.OrderedMessageHandler._pull_message",
            new=MockMessagePuller(
                messages=[
                    {"type": "finish-test", "message_number": 3},
                    {"type": "test", "message_number": 1},
                    {"type": "test", "message_number": 2},
                    {"type": "test", "message_number": 0},
                ]
            ).pull,
        ):
            result = message_handler.handle_messages()

        self.assertEqual(result, "This is the result.")
        self.assertEqual(message_handling_order, [0, 1, 2])

    def test_no_timeout(self):
        """Test that message handling works with no timeout."""
        message_handling_order = []

        message_handler = OrderedMessageHandler(
            subscriber=MockSubscriber(),
            subscription=self.mock_subscription,
            message_handlers={
                "test": self._make_order_recording_message_handler(message_handling_order),
                "finish-test": lambda message: "This is the result.",
            },
        )

        with patch(
            "octue.cloud.pub_sub.service.OrderedMessageHandler._pull_message",
            new=MockMessagePuller(
                messages=[
                    {"type": "finish-test", "message_number": 2},
                    {"type": "test", "message_number": 0},
                    {"type": "test", "message_number": 1},
                ]
            ).pull,
        ):
            result = message_handler.handle_messages(timeout=None)

        self.assertEqual(result, "This is the result.")
        self.assertEqual(message_handling_order, [0, 1])

    def test_error_raised_if_delivery_acknowledgement_not_received_in_time(self):
        """Test that an error is raised if delivery acknowledgement isn't received before the given acknowledgement
        timeout.
        """
        message_handler = OrderedMessageHandler(subscriber=MockSubscriber(), subscription=self.mock_subscription)

        with patch("tests.cloud.pub_sub.mocks.MockSubscriber.pull", return_value=MockPullResponse()):
            with self.assertRaises(exceptions.QuestionNotDelivered):
                message_handler.handle_messages(delivery_acknowledgement_timeout=0)

    def test_delivery_acknowledgement(self):
        """Test that a delivery acknowledgement message is handled correctly."""
        message_handler = OrderedMessageHandler(subscriber=MockSubscriber(), subscription=self.mock_subscription)
        self.assertFalse(message_handler.received_delivery_acknowledgement)

        with patch(
            "octue.cloud.pub_sub.service.OrderedMessageHandler._pull_message",
            new=MockMessagePuller(
                [
                    {
                        "type": "delivery_acknowledgement",
                        "delivery_time": "2021-11-17 17:33:59.717428",
                        "message_number": 0,
                    },
                    {"type": "result", "output_values": None, "output_manifest": None, "message_number": 1},
                ]
            ).pull,
        ):
            result = message_handler.handle_messages()

        self.assertTrue(message_handler.received_delivery_acknowledgement)
        self.assertEqual(result, {"output_values": None, "output_manifest": None})



from octue.resources.label import LabelSet


class Labelable:
    """A mixin class allowing objects to be labelled."""

    def __init__(self, *args, labels=None, **kwargs):
        self.labels = labels
        super().__init__(*args, **kwargs)

    def add_labels(self, *args):
        """Add one or more new labels to the object. New labels will be cleaned and validated."""
        self.labels.update(*args)

    @property
    def labels(self):
        """Get the labels of the labelled object.

        :return iter:
        """
        return self._labels

    @labels.setter
    def labels(self, labels):
        """Overwrite any existing label set and assign new labels."""
        self._labels = LabelSet(labels)



import copy
import json
import os
import tempfile

from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.resources import Datafile, Dataset, Manifest
from tests import TEST_BUCKET_NAME
from tests.base import BaseTestCase
from tests.resources import create_dataset_with_two_files


class TestManifest(BaseTestCase):
    def test_hash_value(self):
        """Test hashing a manifest with multiple datasets gives a hash of length 8."""
        manifest = self.create_valid_manifest()
        hash_ = manifest.hash_value
        self.assertTrue(isinstance(hash_, str))
        self.assertTrue(len(hash_) == 8)

    def test_hashes_for_the_same_manifest_are_the_same(self):
        """Ensure the hashes for two manifests that are exactly the same are the same."""
        first_manifest = self.create_valid_manifest()
        second_manifest = copy.deepcopy(first_manifest)
        self.assertEqual(first_manifest.hash_value, second_manifest.hash_value)

    def test_all_datasets_are_in_cloud(self):
        """Test whether all files of all datasets in a manifest are in the cloud or not can be determined."""
        self.assertFalse(self.create_valid_manifest().all_datasets_are_in_cloud)
        self.assertTrue(Manifest().all_datasets_are_in_cloud)

        files = [
            Datafile(path="gs://hello/file.txt", hypothetical=True),
            Datafile(path="gs://goodbye/file.csv", hypothetical=True),
        ]

        manifest = Manifest(datasets={"my_dataset": Dataset(files=files)})
        self.assertTrue(manifest.all_datasets_are_in_cloud)

    def test_deserialise(self):
        """Test that manifests can be deserialised."""
        manifest = self.create_valid_manifest()
        serialised_manifest = manifest.to_primitive()
        deserialised_manifest = Manifest.deserialise(serialised_manifest)

        self.assertEqual(manifest.name, deserialised_manifest.name)
        self.assertEqual(manifest.id, deserialised_manifest.id)
        self.assertEqual(manifest.absolute_path, deserialised_manifest.absolute_path)

        for key in manifest.datasets.keys():
            self.assertEqual(manifest.datasets[key].name, deserialised_manifest.datasets[key].name)
            self.assertEqual(manifest.datasets[key].id, deserialised_manifest.datasets[key].id)
            self.assertEqual(manifest.datasets[key].absolute_path, deserialised_manifest.datasets[key].absolute_path)

    def test_to_cloud(self):
        """Test that a manifest can be uploaded to the cloud as a serialised JSON file of the Manifest instance via
        (`bucket_name`, `output_directory`) and via `gs_path`.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            dataset = create_dataset_with_two_files(temporary_directory)
            dataset.to_cloud(cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, "my-small-dataset"))

            manifest = Manifest(datasets={"my-dataset": dataset})

            path_to_manifest_file = "manifest.json"
            gs_path = storage.path.generate_gs_path(TEST_BUCKET_NAME, path_to_manifest_file)

            for location_parameters in (
                {"bucket_name": TEST_BUCKET_NAME, "path_to_manifest_file": path_to_manifest_file, "cloud_path": None},
                {"bucket_name": None, "path_to_manifest_file": None, "cloud_path": gs_path},
            ):
                with self.subTest(**location_parameters):
                    manifest.to_cloud(**location_parameters)

                    persisted_manifest = json.loads(
                        GoogleCloudStorageClient().download_as_string(
                            bucket_name=TEST_BUCKET_NAME,
                            path_in_bucket="manifest.json",
                        )
                    )

                    self.assertEqual(
                        persisted_manifest["datasets"]["my-dataset"],
                        "gs://octue-test-bucket/my-small-dataset",
                    )

    def test_to_cloud_without_storing_datasets(self):
        """Test that a manifest can be uploaded to the cloud as a serialised JSON file of the Manifest instance."""
        with tempfile.TemporaryDirectory() as temporary_directory:
            dataset = create_dataset_with_two_files(temporary_directory)
            manifest = Manifest(datasets={"my-dataset": dataset})

            manifest.to_cloud(
                bucket_name=TEST_BUCKET_NAME,
                path_to_manifest_file=storage.path.join("my-manifests", "manifest.json"),
                store_datasets=False,
            )

        persisted_manifest = json.loads(
            GoogleCloudStorageClient().download_as_string(
                bucket_name=TEST_BUCKET_NAME,
                path_in_bucket=storage.path.join("my-manifests", "manifest.json"),
            )
        )

        self.assertEqual(persisted_manifest["datasets"]["my-dataset"], temporary_directory)

    def test_from_cloud(self):
        """Test that a Manifest can be instantiated from the cloud via (`bucket_name`, `output_directory`) and via
        `gs_path`.
        """
        with tempfile.TemporaryDirectory() as temporary_directory:
            dataset = create_dataset_with_two_files(temporary_directory)
            dataset_path = storage.path.generate_gs_path(TEST_BUCKET_NAME, "my_nice_dataset")
            dataset.to_cloud(cloud_path=dataset_path)

            manifest = Manifest(datasets={"my-dataset": dataset})

            manifest.to_cloud(
                bucket_name=TEST_BUCKET_NAME,
                path_to_manifest_file=storage.path.join("my-directory", "manifest.json"),
            )

            path_to_manifest_file = storage.path.join("my-directory", "manifest.json")
            gs_path = storage.path.generate_gs_path(TEST_BUCKET_NAME, path_to_manifest_file)

            for location_parameters in (
                {"bucket_name": TEST_BUCKET_NAME, "path_to_manifest_file": path_to_manifest_file, "cloud_path": None},
                {"bucket_name": None, "path_to_manifest_file": None, "cloud_path": gs_path},
            ):
                with self.subTest(**location_parameters):
                    persisted_manifest = Manifest.from_cloud(**location_parameters)

                    self.assertEqual(persisted_manifest.path, f"gs://{TEST_BUCKET_NAME}/my-directory/manifest.json")
                    self.assertEqual(persisted_manifest.id, manifest.id)
                    self.assertEqual(persisted_manifest.hash_value, manifest.hash_value)
                    self.assertEqual(
                        {dataset.name for dataset in persisted_manifest.datasets.values()},
                        {dataset.name for dataset in manifest.datasets.values()},
                    )

                    for dataset in persisted_manifest.datasets.values():
                        self.assertEqual(dataset.path, dataset_path)
                        self.assertTrue(len(dataset.files), 2)
                        self.assertTrue(all(isinstance(file, Datafile) for file in dataset.files))

    def test_instantiating_from_serialised_cloud_datasets_with_no_dataset_json_file(self):
        """Test that a Manifest can be instantiated from a serialized cloud dataset with no `dataset.json` file. This
        simulates what happens when such a cloud dataset is referred to in a manifest received by a child service.
        """
        GoogleCloudStorageClient().upload_from_string(
            "[1, 2, 3]",
            cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, "my_dataset", "file_0.txt"),
        )

        GoogleCloudStorageClient().upload_from_string(
            "[4, 5, 6]",
            cloud_path=storage.path.generate_gs_path(TEST_BUCKET_NAME, "my_dataset", "file_1.txt"),
        )

        serialised_cloud_dataset = Dataset.from_cloud(cloud_path=f"gs://{TEST_BUCKET_NAME}/my_dataset").to_primitive()

        manifest = Manifest(datasets={"my_dataset": serialised_cloud_dataset})
        self.assertEqual(len(manifest.datasets), 1)
        self.assertEqual(manifest.datasets["my_dataset"].path, f"gs://{TEST_BUCKET_NAME}/my_dataset")
        self.assertEqual(len(manifest.datasets["my_dataset"].files), 2)

    def test_instantiating_from_datasets_from_different_cloud_buckets(self):
        """Test instantiating a manifest from multiple datasets from different cloud buckets."""
        storage_client = GoogleCloudStorageClient()
        storage_client.create_bucket(name="another-test-bucket")

        storage_client.upload_from_string(
            "[1, 2, 3]",
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket="my_dataset_1/file_0.txt",
        )

        storage_client.upload_from_string(
            "[4, 5, 6]",
            bucket_name="another-test-bucket",
            path_in_bucket="my_dataset_2/the_data.txt",
        )

        manifest = Manifest(
            datasets={
                "my_dataset_1": f"gs://{TEST_BUCKET_NAME}/my_dataset_1",
                "my_dataset_2": "gs://another-test-bucket/my_dataset_2",
            }
        )

        self.assertEqual({dataset.name for dataset in manifest.datasets.values()}, {"my_dataset_1", "my_dataset_2"})

        files = [list(dataset.files)[0] for dataset in manifest.datasets.values()]
        self.assertEqual({file.bucket_name for file in files}, {TEST_BUCKET_NAME, "another-test-bucket"})

    def test_instantiating_from_multiple_local_datasets(self):
        """Test instantiating a manifest from multiple local datasets."""
        manifest = Manifest(
            datasets={
                "dataset_0": os.path.join("path", "to", "dataset_0"),
                "dataset_1": os.path.join("path", "to", "dataset_1"),
            },
        )

        self.assertEqual({dataset.name for dataset in manifest.datasets.values()}, {"dataset_0", "dataset_1"})

    def test_deprecation_warning_issued_if_datasets_provided_as_list(self):
        """Test that, if datasets are provided as a list (the old format), a deprecation warning is issued and the list
        is converted to a dictionary (the new format).
        """
        storage_client = GoogleCloudStorageClient()

        storage_client.upload_from_string(
            "[1, 2, 3]",
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket="my_dataset_1/file_0.txt",
        )

        storage_client.upload_from_string(
            "[4, 5, 6]",
            bucket_name=TEST_BUCKET_NAME,
            path_in_bucket="my_dataset_2/the_data.txt",
        )

        with self.assertWarns(DeprecationWarning):
            manifest = Manifest(
                datasets=[
                    f"gs://{TEST_BUCKET_NAME}/my_dataset_1",
                    f"gs://{TEST_BUCKET_NAME}/my_dataset_2",
                ],
                keys={"my_dataset_1": 0, "my_dataset_2": 1},
            )

        self.assertEqual(set(manifest.datasets.keys()), {"dataset_0", "dataset_1"})



import os
import tempfile
from unittest import TestCase, mock

from octue.cloud.deployment.google.answer_pub_sub_question import answer_question
from octue.exceptions import MissingServiceID
from tests.cloud.pub_sub.mocks import MockTopic


SERVICE_ID = "octue.services.14b124f3-8ca9-4baf-99f5-0b79179f04a6"


class TestAnswerQuestion(TestCase):
    def test_error_raised_when_no_service_id_environment_variable(self):
        """Test that a MissingServiceID error is raised if the SERVICE_ID environment variable is missing."""
        with self.assertRaises(MissingServiceID):
            answer_question(
                question={
                    "data": {},
                    "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                },
                project_name="a-project-name",
                credentials_environment_variable="GOOGLE_APPLICATION_CREDENTIALS",
            )

    def test_error_raised_when_service_id_environment_variable_is_empty(self):
        """Test that a MissingServiceID error is raised if the SERVICE_ID environment variable is empty."""
        with mock.patch.dict(os.environ, {"SERVICE_ID": ""}):
            with self.assertRaises(MissingServiceID):
                answer_question(
                    question={
                        "data": {},
                        "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                    },
                    project_name="a-project-name",
                    credentials_environment_variable="GOOGLE_APPLICATION_CREDENTIALS",
                )

    def test_with_no_deployment_configuration_file(self):
        """Test that the Cloud Run `answer_question` function uses the default deployment values when a deployment
        configuration file is not provided.
        """
        with mock.patch.dict(os.environ, {"SERVICE_ID": SERVICE_ID}):
            with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Runner") as mock_runner:
                with mock.patch("octue.cloud.pub_sub.service.Topic", new=MockTopic):
                    with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Service"):
                        answer_question(
                            question={
                                "data": {},
                                "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                            },
                            project_name="a-project-name",
                            credentials_environment_variable="GOOGLE_APPLICATION_CREDENTIALS",
                        )

        mock_runner.assert_called_with(
            **{
                "app_src": ".",
                "twine": "twine.json",
                "configuration_values": None,
                "configuration_manifest": None,
                "output_manifest_path": None,
                "children": None,
                "skip_checks": False,
                "project_name": "a-project-name",
            }
        )

    def test_with_deployment_configuration_file(self):
        """Test that the Cloud Run `answer_question` function uses the values in the deployment configuration file if it
        is provided.
        """
        with mock.patch.dict(
            os.environ,
            {"DEPLOYMENT_CONFIGURATION_PATH": tempfile.NamedTemporaryFile().name, "SERVICE_ID": SERVICE_ID},
        ):
            with mock.patch(
                "octue.cloud.deployment.google.answer_pub_sub_question._get_deployment_configuration",
                return_value={"app_dir": "/path/to/app_dir"},
            ):
                with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Runner") as mock_runner:
                    with mock.patch("octue.cloud.pub_sub.service.Topic", new=MockTopic):
                        with mock.patch("octue.cloud.deployment.google.answer_pub_sub_question.Service"):
                            answer_question(
                                question={
                                    "data": {},
                                    "attributes": {"question_uuid": "8c859f87-b594-4297-883f-cd1c7718ef29"},
                                },
                                project_name="a-project-name",
                                credentials_environment_variable="GOOGLE_APPLICATION_CREDENTIALS",
                            )

        mock_runner.assert_called_with(
            **{
                "app_src": "/path/to/app_dir",
                "twine": "twine.json",
                "configuration_values": None,
                "configuration_manifest": None,
                "output_manifest_path": None,
                "children": None,
                "skip_checks": False,
                "project_name": "a-project-name",
            }
        )



from octue.mixins import CoolNameable
from tests.base import BaseTestCase


class NamedThing(CoolNameable):
    def __init__(self, name, *args, **kwargs):
        self.name = name
        super().__init__(*args, **kwargs)


class PotentiallyNamedThing(CoolNameable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = kwargs.pop("name", None)


class TestCoolNameable(BaseTestCase):
    def test_name_has_two_components(self):
        """Test that the cool name is in the format of <word>-<word>."""
        cool_nameable = CoolNameable()
        self.assertEqual(len(cool_nameable.name.split("-")), 2)

    def test_names_are_different(self):
        """Test that the cool names of two different instances are different."""
        cool_nameables = {CoolNameable().name for _ in range(5)}
        self.assertEqual(len(cool_nameables), 5)

    def test_cool_name_is_not_used_if_subclass_already_has_name_attribute(self):
        """Ensure that CoolNameable does not override the existing name attribute of a subclass instance."""
        named_thing = NamedThing(name="I have a name")
        self.assertEqual(named_thing.name, "I have a name")

    def test_cool_name_is_not_used_if_subclass_already_has_name_in_kwargs(self):
        """Ensure that CoolNameable does not override the existing name keyword argument of a subclass instance."""
        named_thing = PotentiallyNamedThing(name="I have a name")
        self.assertEqual(named_thing.name, "I have a name")

    def test_cool_name_is_applied_when_existing_name_is_none_in_subclass(self):
        """Test that a cool name is applied if the existing name attribute of the subclass is set to None."""
        named_thing = NamedThing(name=None)
        self.assertTrue(named_thing is not None)

    def test_cool_name_is_applied_when_no_name_kwarg_given_to_subclass(self):
        """Test that a cool name is applied if there is no name keyword argument is given to the subclass."""
        named_thing = PotentiallyNamedThing()
        self.assertTrue(named_thing is not None)



import logging
import os

import apache_beam
from apache_beam.options.pipeline_options import PipelineOptions
from apache_beam.runners.dataflow.dataflow_runner import DataflowRunner
from apache_beam.runners.dataflow.internal.apiclient import DataflowJobAlreadyExistsError

from octue import REPOSITORY_ROOT
from octue.cloud.deployment.google.answer_pub_sub_question import answer_question
from octue.cloud.pub_sub import Topic
from octue.cloud.pub_sub.service import Service
from octue.exceptions import DeploymentError
from octue.resources.service_backends import GCPPubSubBackend


logger = logging.getLogger(__name__)


DEFAULT_DATAFLOW_TEMPORARY_FILES_LOCATION = "gs://octue-sdk-python-dataflow-temporary-data/temp"
DEFAULT_SETUP_FILE_PATH = os.path.join(REPOSITORY_ROOT, "setup.py")


def create_streaming_job(
    service_name,
    service_id,
    project_name,
    region,
    image_uri,
    setup_file_path=DEFAULT_SETUP_FILE_PATH,
    temporary_files_location=DEFAULT_DATAFLOW_TEMPORARY_FILES_LOCATION,
    service_account_email=None,
    worker_machine_type=None,
    maximum_instances=None,
    update=False,
    extra_options=None,
):
    """Deploy an `octue` service as a streaming Google Dataflow Prime job.

    :param str service_name: the name to give the Dataflow job
    :param str service_id: the Pub/Sub topic name for the Dataflow job to subscribe to
    :param str project_name: the name of the project to deploy the job to
    :param str region: the region to deploy the job to
    :param str image_uri: the URI of the `apache-beam`-based Docker image to use for the job
    :param str setup_file_path: path to the python `setup.py` file to use for the job
    :param str temporary_files_location: a Google Cloud Storage path to save temporary files from the job at
    :param str|None service_account_email: the email of the service account to run the Dataflow VMs as
    :param str|None worker_machine_type: the machine type to create Dataflow worker VMs as. See https://cloud.google.com/compute/docs/machine-types for a list of valid options. If not set, the Dataflow service will choose a reasonable default.
    :param int|None maximum_instances: the maximum number of workers to use when executing the Dataflow job
    :param bool update: if `True`, update the existing job with the same name
    :param dict|None extra_options: any further arguments to be passed to Apache Beam as pipeline options
    :raise DeploymentError: if a Dataflow job with the service name already exists
    :return None:
    """
    pipeline_options = {
        "project": project_name,
        "region": region,
        "temp_location": temporary_files_location,
        "job_name": service_name,
        "sdk_container_image": image_uri,
        "setup_file": os.path.abspath(setup_file_path),
        "update": update,
        "streaming": True,
        **(extra_options or {}),
    }

    if service_account_email:
        pipeline_options["service_account_email"] = service_account_email

    if worker_machine_type:
        pipeline_options["worker_machine_type"] = worker_machine_type
    else:
        # Dataflow Prime can only be used if a worker machine type is not specified.
        pipeline_options["dataflow_service_options"] = ["enable_prime"]

    if maximum_instances:
        pipeline_options["max_num_workers"] = maximum_instances

    pipeline_options = PipelineOptions.from_dictionary(pipeline_options)
    pipeline = apache_beam.Pipeline(options=pipeline_options)

    service_topic = Topic(name=service_id, service=Service(backend=GCPPubSubBackend(project_name=project_name)))
    service_topic.create(allow_existing=True)

    (
        pipeline
        | "Read from Pub/Sub" >> apache_beam.io.ReadFromPubSub(topic=service_topic.path, with_attributes=True)
        | "Answer question" >> apache_beam.Map(lambda question: answer_question(question, project_name=project_name))
    )

    try:
        DataflowRunner().run_pipeline(pipeline, options=pipeline_options)
    except DataflowJobAlreadyExistsError:
        raise DeploymentError(f"A Dataflow job with name {service_name!r} already exists.") from None



import json
import logging
from logging import makeLogRecord
from unittest.mock import patch

from octue.cloud.pub_sub.logging import GooglePubSubHandler
from octue.resources.service_backends import GCPPubSubBackend
from tests.base import BaseTestCase
from tests.cloud.pub_sub.mocks import MESSAGES, MockService, MockTopic


class NonJSONSerialisable:
    def __repr__(self):
        return "NonJSONSerialisableInstance"


class TestGooglePubSubHandler(BaseTestCase):
    def test_emit(self):
        """Test the log message is published when `GooglePubSubHandler.emit` is called."""
        backend = GCPPubSubBackend(project_name="blah")
        service = MockService(backend=backend)
        topic = MockTopic(name="world", namespace="hello", service=service)
        topic.create()

        log_record = makeLogRecord({"msg": "Starting analysis."})
        GooglePubSubHandler(service.publisher, topic, "analysis-id").emit(log_record)

        self.assertEqual(json.loads(MESSAGES[topic.name][0].data.decode())["log_record"]["msg"], "Starting analysis.")

    def test_emit_with_non_json_serialisable_args(self):
        """Test that non-JSON-serialisable arguments to log messages are converted to their string representation
        before being serialised and published to the Pub/Sub topic.
        """
        backend = GCPPubSubBackend(project_name="blah")
        service = MockService(backend=backend)
        topic = MockTopic(name="world-1", namespace="hello", service=service)
        topic.create()

        non_json_serialisable_thing = NonJSONSerialisable()

        # Check that it can't be serialised to JSON.
        with self.assertRaises(TypeError):
            json.dumps(non_json_serialisable_thing)

        record = logging.makeLogRecord(
            {"msg": "%r is not JSON-serialisable but can go into a log message", "args": (non_json_serialisable_thing,)}
        )

        with patch("tests.cloud.pub_sub.mocks.MockPublisher.publish") as mock_publish:
            GooglePubSubHandler(service.publisher, topic, "analysis-id").emit(record)

        self.assertEqual(
            json.loads(mock_publish.call_args.kwargs["data"].decode())["log_record"]["msg"],
            "NonJSONSerialisableInstance is not JSON-serialisable but can go into a log message",
        )



from unittest import TestCase

from octue import exceptions
from octue.resources.tag import TagDict


class TestTagDict(TestCase):
    def test_instantiate_from_dict(self):
        """Test that a TagDict can be instantiated from a dictionary."""
        tag_dict = TagDict({"a": 1, "b": 2})
        self.assertEqual(tag_dict, {"a": 1, "b": 2})

    def test_instantiate_from_kwargs(self):
        """Test that a TagDict can be instantiated from kwargs."""
        tag_dict = TagDict(**{"a": 1, "b": 2})
        self.assertEqual(tag_dict, {"a": 1, "b": 2})

    def test_instantiation_fails_if_tag_name_fails_validation(self):
        """Test that TagDict instantiation fails if any keys don't conform to the tag name pattern."""
        with self.assertRaises(exceptions.InvalidTagException):
            TagDict({".blah.": "blue"})

    def test_update_fails_if_tag_name_fails_validation(self):
        """Test that updating fails if any keys don't conform to the tag name pattern."""
        tag_dict = TagDict({"a": 1, "b": 2})

        with self.assertRaises(exceptions.InvalidTagException):
            tag_dict.update({"@": 3, "d": 4})

        self.assertEqual(tag_dict, {"a": 1, "b": 2})

    def test_update(self):
        """Test that TagDicts can be updated with tags with valid names."""
        tag_dict = TagDict({"a": 1, "b": 2})
        tag_dict.update({"c": 3, "d": 4})
        self.assertEqual(tag_dict, {"a": 1, "b": 2, "c": 3, "d": 4})

    def test_setitem_fails_if_tag_name_fails_validation(self):
        """Test that setting an item on a TagDict fails if the name fails validation."""
        tag_dict = TagDict()

        with self.assertRaises(exceptions.InvalidTagException):
            tag_dict["@@@"] = 9

    def test_setitem(self):
        """Test setting an item on a TagDict."""
        tag_dict = TagDict()
        tag_dict["hello"] = 9
        self.assertEqual(tag_dict, {"hello": 9})

    def test_equality_to_dict(self):
        """Test that TagDicts compare equal to dictionaries with the same contents."""
        tag_dict = TagDict({"a": 1, "b": 2})
        self.assertEqual(tag_dict, {"a": 1, "b": 2})
